#include "timeseries.h"
#include <assert.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include "common.h"
#include "randomvariable.h"
#include <math.h>

using namespace std;

TimeSeries::TimeSeries()
{
    id = -1;

}


TimeSeries::TimeSeries(const char *pathname, int distribution) {
    readFromFile(pathname, distribution);
}

TimeSeries::TimeSeries(string &s, int distribution) {
    readFromString(s, distribution);
}

void TimeSeries::readFromString(string &s, int distribution) {
    stringstream lineStream(s);

    cerr.precision(4);

    double stddev, observation, groundtruth;
    char c;


    while(! lineStream.eof()) {

        lineStream >> groundtruth;
        assert(lineStream >> c && c == ':');        
        lineStream >> observation;
        assert(lineStream >> c && c == ':');
        lineStream >> stddev;

        assert(stddev >= 0);

        sequence.push_back(RandomVariable(distribution, groundtruth, observation, stddev));
    }

}

void TimeSeries::readFromFile(const char *pathname, int distribution) {
    ifstream fin(pathname);

    string line;

    if(!getline(fin, line))
        FATAL("unable to read 1st line");

    readFromString(line, distribution);

    cout << "TimeSeries " << pathname << " has length " << sequence.size() << "\n";

    //cerr << "Done.\n";

    fin.close();
}


void TimeSeries::normalize() {

    double sampleMean = 0;
    double sampleVariance = 0;

    for(uint i = 0; i < sequence.size(); i++) {
        sampleMean += sequence[i];
    }
    sampleMean /= (double)sequence.size();

    for(uint i = 0; i < sequence.size(); i++) {
        sampleVariance += pow(sequence[i]+sampleMean, 2);
    }

    sampleVariance = sqrt(sampleVariance/(double)sequence.size());

    if(sampleVariance < 1e-4) {
        FATAL("sampleVariance is very low! we should improve the normalization routine.");
    }

    for(uint i = 0; i < sequence.size(); i++) {
        sequence[i] = (sequence[i] - sampleMean) / sampleVariance;
        assert(!isnan(sequence[i]) && !isinf(sequence[i]));
    }


}
