/*
 * GroundTruthCalculator.cpp
 *
 *  Created on: May 9, 2011
 *      Author: besa
 */

#include "GroundTruthCalculator.h"
#include "Benchmark.h"

#include <string>
#include <iostream>
#include <fstream>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
using namespace std;
using namespace boost;

GroundTruthCalculator::GroundTruthCalculator() {
	GROUND_TRUTH_PATH = "./data/groundtruth";
	RESULT_PATH = "./data/euclidean";
}

GroundTruthCalculator::~GroundTruthCalculator() {
	// TODO Auto-generated destructor stub
}

void GroundTruthCalculator::read()
{
	ifstream seriesfile (GROUND_TRUTH_PATH.c_str());
	string line;
	if (seriesfile.is_open())
	{
		while ( seriesfile.good() )
		{
			  getline (seriesfile,line);
			  if (line.length() == 0)
				  continue;
			  TimeSeries newSeries;
			  vector<string> values;
			  boost::split(values, line, boost::is_any_of(" "));
                          for (unsigned int i = 0; i < values.size(); i++)
			  {
				  newSeries.push_back(lexical_cast<double>(values[i]));
			  }

			  this->allSeries.push_back(newSeries);
		}
	}
	seriesfile.close();
	//normalize the series
	for(unsigned int i = 0; i < this->allSeries.size(); i++)
		this->allSeries[i] = this->normalize(this->allSeries[i]);
}

void GroundTruthCalculator::write()
{
	ofstream result_file;
	result_file.open (RESULT_PATH.c_str());
	for(unsigned int i = 0; i < this->allSeries.size(); i++)
	{
		TimeSeries t1 = this->allSeries[i];
		for (unsigned int j = 0; j < this->allSeries.size(); j++)
		{
			TimeSeries t2 = this->allSeries[j];
			double distance = this->euclidean(t1, t2);
			result_file << distance * distance << " ";
		}
		result_file << endl;
	}
	result_file.close();
}

double GroundTruthCalculator::euclidean(TimeSeries t1, TimeSeries t2)
{
	double sum = 0.0;
	for (unsigned int i = 0; i < t1.size(); i++)
	{
		double diff = t1[i] - t2[i];
		sum += diff * diff;
	}
	return sqrt(sum);
}

double GroundTruthCalculator::length(TimeSeries t)
{
	double sum = 0.0;
	for (unsigned int i = 0; i < t.size(); i++)
		sum += t[i] * t[i];
	return sqrt(sum);
}

GroundTruthCalculator::TimeSeries GroundTruthCalculator::normalize(TimeSeries sequence)
{
	double sampleMean = 0;
	double sampleVariance = 0;
	for(unsigned int i = 0; i < sequence.size(); i++) {
		sampleMean += sequence[i];
	}
	sampleMean /= (double)sequence.size();

	for(unsigned int i = 0; i < sequence.size(); i++) {
		sampleVariance += pow(sequence[i] - sampleMean, 2);
	}
	sampleVariance = sqrt(sampleVariance/(double)sequence.size());
	if(sampleVariance < 1e-4) {
		cout << "sampleVariance is very low! we should improve the normalization routine.";
		exit(1);
	}
	for(unsigned int i = 0; i < sequence.size(); i++) {
		sequence[i] = (sequence[i] - sampleMean) / sampleVariance;
	}
	return sequence;
}
