/*
 * PrecisionRecall.cpp
 *
 *  Created on: May 14, 2011
 *      Author: besa
 */

#include "PrecisionRecall.h"

#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <vector>

using namespace boost;

PrecisionRecall::PrecisionRecall(int n, double r_value,
		string resultPath, string groundTruth) {
	this->RESULT_PATH = resultPath;
	this->GROUND_TRUTH_EUCLIDEAN = groundTruth;
	this->noSeries = n;
	this->r = r_value;
	this->precision = 0;
	this->recall = 0;
	this->f1 = 0;
}

PrecisionRecall::PrecisionRecall(){
}

PrecisionRecall::~PrecisionRecall() {
	// TODO Auto-generated destructor stub
}

void PrecisionRecall::calculate()
{
	this->read_ground_truth();
//	string output_filename = this->OUTPUT_PATH;
//	ofstream pr_file;
//	pr_file.open (output_filename.c_str());
//	pr_file << "TS" << "\t\t" <<"CORR" << "\t\t" << "ERR" << "\t\t" << "N_R" << "\t\t" << "TOTAL";
//	pr_file << "\t\t" << "PREC" << "\t\t\t" << "RECALL" << "\t\t\t" << "F1" << endl;
	int t = 0;
	int c = 0;
	int e = 0;
	int nr = 0;
	double macro_p = 0;
	double macro_r = 0;
	double macro_f1 = 0;
	for(int i = 0; i < this->noSeries; i++)
	{
		string filename = this->RESULT_PATH + boost::lexical_cast<std::string>(i) + ".dat";
		fstream resultsfile (filename.c_str());
		string line;
		int total = 0;
		int correct = 0;
		int error = 0;
		int not_retrieved;
		if (resultsfile.is_open())
		{
			while ( resultsfile.good() )
			{
				getline (resultsfile,line);
				if (line.length() == 0)
					continue;
				try
				{
					int series = lexical_cast<int>(line);
					bool in = false;
					for(unsigned int j = 0; j < this->euclidean_matrix[i].size(); j++)
					{
						if (this->euclidean_matrix[i][j] == series)
						{
							in = true;
							break;
						}
					}
					if (in == true)
						correct++;
					else
						error++;
					total++;
				}
				catch( boost::bad_lexical_cast const& )
				{
					cout << line << "casting data to int error" << endl;
				}
			}
		}

		not_retrieved = this->euclidean_matrix[i].size() - correct;
//		pr_file << i << "\t\t" << correct << "\t\t" << error << "\t\t" << not_retrieved << "\t\t" << total;
		double precision = (double) correct / ((double) correct + (double) error);
		double recall = (double) correct / ((double) correct + (double) not_retrieved);
		double f1 = 2 * precision * recall / (precision + recall);
//		pr_file << "\t\t" << precision << "\t\t" << recall << "\t\t" << f1;
//		pr_file << endl;
		t += total;
		c += correct;
		e += error;
		nr += not_retrieved;
		macro_p += precision;
		macro_r += recall;
		macro_f1 += f1;
	}
//	pr_file << "--------------------------------------------------------------------------------------";
//	pr_file << endl;
        //double micro_p = (double) c / ((double) c + (double) e);
        //double micro_r = (double) c / ((double) c + (double) nr);
        //double micro_f1 = 2 * micro_p * micro_r / (micro_p + micro_r); //bymichele: not used!
//	pr_file << "TOTAL MICRO" << "\t\t" << c << "\t\t" << e << "\t\t" << nr << "\t\t" << t;
//	pr_file << "\t\t" << micro_p << "\t\t" << micro_r << "\t\t" << micro_f1;
//	pr_file << endl;
	macro_p = macro_p / this->noSeries;
	macro_r = macro_r / this->noSeries;
	macro_f1 = macro_f1 / this->noSeries;
//	pr_file << "TOTAL MACRO" << "\t\t" << c << "\t\t" << e << "\t\t" << nr << "\t\t" << t;
//		pr_file << "\t\t" << macro_p << "\t\t" << macro_r << "\t\t" << macro_f1;
//	pr_file.close();
	this->precision = macro_p;
	this->recall = macro_r;
	this->f1 = macro_f1;
}

void PrecisionRecall::read_ground_truth()
{
	string filename = GROUND_TRUTH_EUCLIDEAN;
	ifstream seriesfile (filename.c_str());
	string line;
	if (seriesfile.is_open())
	{
		while ( seriesfile.good() )
		{
			  getline (seriesfile,line);
			  if (line.length() == 0)
				  continue;
			  vector<int> results_ts;
			  vector<string> data;
			  boost::split(data, line, boost::is_any_of(" "));
			  for(unsigned int k = 0; k < data.size() - 1; k++)
			  {
				  try
				  {
					  double distance = lexical_cast<double>(data[k]);
					  if (distance <= this->r)
					  {
						  results_ts.push_back(k);
					  }
				  }
				  catch( boost::bad_lexical_cast const& )
				  {
					  cout << data[k] << "casting data to double error" << endl;
				  }
			  }
			  this->euclidean_matrix.push_back(results_ts);
		}
	}
	seriesfile.close();
}

void PrecisionRecall::average()
{
	vector<PrecisionVector> all_distributions;
	vector<string> datasets = this->initDatasets();
	vector<string> distributions;
	distributions.push_back("uniform");
	distributions.push_back("normal");
	distributions.push_back("exponential");
	for(unsigned int ds = 0; ds < distributions.size(); ds++)
	{
		PrecisionVector one_distribution;
		for(int i = 0; i < 10; i++)
		{
			Set s;
			s.F1 = 0;
			s.Precision = 0;
			s.Recall = 0;
			one_distribution.push_back(s);
		}
		all_distributions.push_back(one_distribution);
	}

	for(unsigned int d = 0; d < datasets.size(); d++)
	{
		for(unsigned int ds = 0; ds < distributions.size(); ds++)
		{
			string in_filename = "./data/graphs/All 0.95/" + datasets[d] + "_proud_" + distributions[ds] + ".dat";
			fstream resultsfile (in_filename.c_str());
			string line_in;
			if (resultsfile.is_open())
			{
				int v = 0;
				while ( resultsfile.good() )
				{
					getline (resultsfile,line_in);
					if (line_in.length() == 0)
						continue;
					vector<string> data;
					boost::split(data, line_in, boost::is_any_of(" "));
					all_distributions[ds][v].Precision += lexical_cast<double>(data[2]);
					all_distributions[ds][v].Recall += lexical_cast<double>(data[3]);
					all_distributions[ds][v].F1 += lexical_cast<double>(data[4]);
					v++;
				}
			}
		}
	}
	for(unsigned int ds = 0; ds < distributions.size(); ds++)
	{
		for(int i = 0; i < 10; i++)
		{
			all_distributions[ds][i].F1 = all_distributions[ds][i].F1 / (double) datasets.size();
			all_distributions[ds][i].Precision = all_distributions[ds][i].Precision / (double) datasets.size();
			all_distributions[ds][i].Recall = all_distributions[ds][i].Recall / (double) datasets.size();
		}
	}
	this->write_average(all_distributions);
}

vector<string> PrecisionRecall::initDatasets()
{
	vector<string> datasets;
	datasets.push_back("50words");
	datasets.push_back("Adiac");
	datasets.push_back("Beef");
	datasets.push_back("CBF");
	datasets.push_back("Coffee");
	datasets.push_back("ECG200");
	datasets.push_back("FaceAll");
	datasets.push_back("FaceFour");
	datasets.push_back("FISH");
	datasets.push_back("Gun_Point");
	datasets.push_back("Lighting2");
	datasets.push_back("Lighting7");
	datasets.push_back("OliveOil");
	datasets.push_back("OSULeaf");
	datasets.push_back("SwedishLeaf");
	datasets.push_back("synthetic_control");
	datasets.push_back("Trace");
	return datasets;
}

void PrecisionRecall::write_average(vector<PrecisionVector> all)
{
	vector<string> distributions;
	distributions.push_back("uniform");
	distributions.push_back("normal");
	distributions.push_back("exponential");
	for(unsigned int ds = 0; ds < all.size(); ds++)
	{
		string out_filename = "./data/graphs/All 0.95/all_" + distributions[ds] + ".dat";
		ofstream all_file;
		all_file.open (out_filename.c_str());
		for(int v = 0; v < 10; v++)
		{
			double var = (v + 1) * 0.2;
			Set set = all[ds][v];
			all_file << var << " " << set.Precision << " " << set.Recall << " " << set.F1 << endl;
		}
		all_file.close();
	}
}
