/*
 * main.cpp
 *
 *  Created on: May 9, 2011
 *      Author: besa
 */
#include <limits>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <iostream>
#include <fstream>
#include <map>
#include <boost/math/special_functions/erf.hpp>
#include <sys/stat.h>
#include <sys/types.h>
#include <boost/algorithm/string.hpp>
#include <sys/stat.h>

#include "main.h"
#include "timeseries.h"
#include "timeseriescollection.h"
#include "common.h"
#include "euclidean.h"
#include "QueryEngine.h"
#include "Benchmark.h"
#include "GroundTruthCalculator.h"
#include "PrecisionRecall.h"
#include "proud.h"
#include "precisionrecallM.h" // pr from michele.

using namespace boost::posix_time;
using namespace boost::gregorian;
using namespace boost;
using namespace std;

int main(int argc, char **argv);
void initOpt(int argc, char **argv);
void help();
int getdir(string dir, vector<string> &files);
map<string, double> readThresholds();
bool FolderExists(string file);
void run_fixed_variance();
void run_fixed_tao();
void average_runtimes();
void average_times();
void find_max_tao();

void michele();
void experiment4(double probabThreshold);
void experiment4_evaluate(TimeSeriesCollection &db, PROUD &proud, vector<PairIdDistance> &topIdDists, PrecisionRecallM &prProud, PrecisionRecallM &prEuclidean, double probabThreshold);
void experiment6(double probabThreshold);

OPT o;

void  initOpt() {
    o.rfileThresholds = "/Users/besa/Thesis/Euclidean/thresholds.txt";
    o.rfileEuclidean = "/Users/besa/Thesis/Euclidean/euclidean_";
}

int main(int argc, char **argv) {

    michele();

    //	GroundTruthCalculator g;
    //	g.calculate();
    string input;
    int answer = 0;
    cout << "Which experiment do you want to run?" << endl;
    cout << "Press 1 for fixed tao experiment OR 2 for fixed variance experiment" << endl;
    getline (cin, input);
    answer = lexical_cast<double>(input);
    if (answer == 1)
    {
        run_fixed_tao();
    }
    else if (answer == 2)
    {
        run_fixed_variance();
    }
}

void average_runtimes()
{
    vector<string> distributions;
    distributions.push_back("uniform");
    distributions.push_back("normal");
    distributions.push_back("exponential");
    string folder = "/Users/besa/Thesis/udatasets/";
    vector<string> folders = vector<string>();
    getdir(folder.c_str(), folders);
    vector< vector<std::pair<PrecisionRecall::Set, double> > > all_values;
    vector<string> tao;
    tao.push_back("0.05");
    tao.push_back("0.15");
    tao.push_back("0.25");
    tao.push_back("0.35");
    tao.push_back("0.45");
    tao.push_back("0.55");
    tao.push_back("0.65");
    tao.push_back("0.75");
    tao.push_back("0.85");
    tao.push_back("0.95");
    for(unsigned int d = 0; d < distributions.size(); d++)
    {
        PrecisionRecall::Set empty;
        empty.F1 = 0;
        empty.Precision = 0;
        empty.Recall = 0;
        vector<std::pair<PrecisionRecall::Set, double> > one_distribution;
        for(unsigned int v = 0; v < tao.size(); v++)
        {
            std::pair<PrecisionRecall::Set, double> pair =
                    std::pair<PrecisionRecall::Set, double> (empty, 0.0);
            one_distribution.push_back(pair);
        }
        all_values.push_back(one_distribution);
    }
    for(unsigned int d = 0; d < distributions.size(); d++)
    {
        for(unsigned int f = 0; f < folders.size(); f++)
        {
            string in_filename = "./data/graphs/" + folders[f] +
                                 "_proud_" + distributions[d] + "_fixed_var.dat";
            fstream resultsfile (in_filename.c_str());
            string line_in;
            if (resultsfile.is_open())
            {
                int v = 0;
                while ( resultsfile.good() )
                {
                    getline (resultsfile,line_in);
                    if (line_in.length() == 0)
                        continue;
                    vector<string> data;
                    boost::split(data, line_in, boost::is_any_of(" "));
                    all_values[d][v].first.Precision += lexical_cast<double>(data[3]);
                    all_values[d][v].first.Recall += lexical_cast<double>(data[4]);
                    all_values[d][v].first.F1 += lexical_cast<double>(data[5]);
                    all_values[d][v].second += lexical_cast<double>(data[6]);
                    v++;
                }
            }
            resultsfile.close();
        }
    }
    //write
    for(unsigned int d = 0; d < distributions.size(); d++)
    {
        string out_filename = "./data/graphs/time_" + distributions[d] + ".dat";
        ofstream pr_file;
        pr_file.open (out_filename.c_str());
        for(unsigned int t = 0; t < all_values[d].size(); t++)
        {
            all_values[d][t].first.Precision = all_values[d][t].first.Precision / folders.size();
            all_values[d][t].first.Recall = all_values[d][t].first.Recall / folders.size();
            all_values[d][t].first.F1 = all_values[d][t].first.F1 / folders.size();

            all_values[d][t].second = all_values[d][t].second / folders.size();
            string line = tao[t] + " " +
                          lexical_cast<std::string>(all_values[d][t].first.Precision) + " " +
                          lexical_cast<std::string>(all_values[d][t].first.Recall) + " " +
                          lexical_cast<std::string>(all_values[d][t].first.F1) + " " +
                          lexical_cast<std::string>(all_values[d][t].second);
            pr_file << line;
            pr_file << endl;
        }
        pr_file.close();
    }
}

void run_fixed_variance()
{
    //get probability threshold from input
    string input;
    cout << "Variance: " << endl;
    getline (cin, input);
    string fixed_var = input;

    //get the location of the files
    cout << "Where are the datasets located?" << endl;
    getline (cin, input);
    string folder = input;

    //get the location of the thresholds' file
    cout << "Where are the thresholds located? (Full Path to the file)" << endl;
    getline (cin, input);
    o.rfileThresholds = input.c_str();

    //get the location of the euclidean distances
    cout << "Where are the files of euclidean distances located? (Full Path to the folder /)" << endl;
    getline (cin, input);
    o.rfileEuclidean = input + "euclidean_";

    //specify the location of output files
    cout << "Where do you want to put the output files? (Full Path to the folder /)" << endl;
    getline (cin, input);
    string output_location = input;

    //specify the location of temporary results files
    cout << "Where do you want to put the temporary result files? (Full Path to the folder /)" << endl;
    getline (cin, input);
    string resultPath = input;

    map<string, double> map_th = readThresholds();
    vector<string> distributions;
    distributions.push_back("uniform");
    distributions.push_back("normal");
    distributions.push_back("exponential");
    vector<string> tao;
    tao.push_back("0.05");
    tao.push_back("0.15");
    tao.push_back("0.25");
    tao.push_back("0.35");
    tao.push_back("0.45");
    tao.push_back("0.55");
    tao.push_back("0.65");
    tao.push_back("0.75");
    tao.push_back("0.85");
    tao.push_back("0.95");
    vector<string> folders = vector<string>();
    getdir(folder.c_str(), folders);
    Benchmark benchmarkall("ALL DATASETS");
    benchmarkall.start(true);
    for(unsigned int d = 0; d < distributions.size(); d++)
    {
        for(unsigned int f = 0; f < folders.size(); f++)
        {
            string base_output_filename = output_location + folders[f] + "_proud_" + distributions[d] + "_fixed_var.dat";
            ofstream pr_file;
            pr_file.open (base_output_filename.c_str());
            string base_input_filename  = folder + folders[f] +"/" +
                                          folders[f] + "_dust_";
            string input_filename  = base_input_filename + distributions[d] + "_" + fixed_var + "_.dat";
            string euclidean_path = o.rfileEuclidean +  folders[f] + ".dat";
            double topn_threshold = map_th.find(folders[f])->second;
            for (unsigned int t = 0; t < tao.size(); t++)
            {
                string line = lexical_cast<std::string> (d) + " " + fixed_var + " " + tao[t] + " ";
                Benchmark benchmark(folders[f] + " dataset | "
                                    + distributions[d] + " distribution | " + fixed_var + " variance | " +
                                    tao[t] + " tao ");
                benchmark.start(true);
                char* cstr = new char[input_filename.length() +1];
                strcpy(cstr, input_filename.c_str());
                o.rfileCollection = cstr;
                TimeSeriesCollection db(o.rfileCollection, 1);
                int length = db.sequences[0].sequence.size();
                int noSeries = db.sequences.size();
                //section parameters
                long int from = 0;
                long int reading_length = length;

                //query the collection
                double milliseconds = 0;
                for(int i = 0; i < noSeries; i++)
                {
                    TimeSeries &query = db.sequences[i];
                    Euclidean euclideanTrue(db, true);
                    Euclidean euclidean(db);
                    euclideanTrue.rangeQuery(query, topn_threshold);
                    double euclideanThreshold = euclidean.distance(query, db.sequences[euclideanTrue.largestDistanceId-1]);
                    QueryEngine qe(db, from, reading_length, lexical_cast<double> (tao[t]), euclideanThreshold, i);
                    qe.ROOT_FOLDER = resultPath;
                    time_duration single = qe.query();
                    milliseconds += single.total_milliseconds();
                }
                milliseconds = milliseconds / noSeries;
                double seconds = (double) milliseconds / (double) 1000;
                //calculate precision recall
                PrecisionRecall pr(noSeries, topn_threshold, resultPath, euclidean_path);
                pr.calculate();
                line += lexical_cast<std::string>(pr.precision) + " ";
                line += lexical_cast<std::string>(pr.recall) + " ";
                line += lexical_cast<std::string>(pr.f1) + " ";
                line += lexical_cast<std::string>(seconds);
                pr_file << line;
                pr_file << endl;
                benchmark.end(true);
            }
            pr_file.close();
        }
    }
    benchmarkall.end(true);
}

void run_fixed_tao()
{
    //get probability threshold from input
    string input;
    double prob_threshold = 0;
    cout << "Probability Threshold: " << endl;
    getline (cin, input);
    prob_threshold = lexical_cast<double>(input);

    //get the location of the files
    cout << "Where are the datasets located?" << endl;
    getline (cin, input);
    string folder = input;

    //get the location of the thresholds' file
    cout << "Where are the thresholds located? (Full Path to the file)" << endl;
    getline (cin, input);
    o.rfileThresholds = input.c_str();

    //get the location of the euclidean distances
    cout << "Where are the files of euclidean distances located? (Full Path to the folder /)" << endl;
    getline (cin, input);
    o.rfileEuclidean = input + "euclidean_";

    //specify the location of output files
    cout << "Where do you want to put the output files? (Full Path to the folder /)" << endl;
    getline (cin, input);
    string output_location = input;

    //specify the location of temporary results files
    cout << "Where do you want to put the temporary result files? (Full Path to the folder /)" << endl;
    getline (cin, input);
    string resultPath = input;


    map<string, double> map_th = readThresholds();
    vector<string> distributions;
    distributions.push_back("uniform");
    distributions.push_back("normal");
    distributions.push_back("exponential");
    vector<string> variances;
    variances.push_back("0.0");
    variances.push_back("0.2");
    variances.push_back("0.4");
    variances.push_back("0.6");
    variances.push_back("0.8");
    variances.push_back("1.0");
    variances.push_back("1.2");
    variances.push_back("1.4");
    variances.push_back("1.6");
    variances.push_back("1.8");
    variances.push_back("2.0");
    vector<string> folders = vector<string>();
    getdir(folder.c_str(), folders);
    Benchmark benchmarkall("ALL DATASETS");
    benchmarkall.start(true);
    for(unsigned int f = 0; f < folders.size(); f++)
    {
        string base_output_filename = output_location + folders[f] + "_proud_";
        string base_input_filename  = folder + folders[f] +"/" +
                                      folders[f] + "_dust_";
        double topn_threshold = map_th.find(folders[f])->second;
        string euclidean_path = o.rfileEuclidean +  folders[f] + ".dat";
        for(unsigned int d = 0; d < distributions.size(); d++)
        {
            string output_filename = base_output_filename + distributions[d] + ".dat";
            ofstream pr_file;
            pr_file.open (output_filename.c_str());
            for(unsigned int v = 0; v < variances.size(); v++)
            {
                string line = lexical_cast<std::string>(d) + " " + variances[v] + " ";
                Benchmark benchmark(folders[f] + " dataset | " + distributions[d] + " distribution | " + variances[v] + " variance");
                benchmark.start(true);
                string input_filename  = base_input_filename + distributions[d] + "_" + variances[v] + "_.dat";
                char* cstr = new char[input_filename.length() +1];
                strcpy(cstr, input_filename.c_str());
                o.rfileCollection = cstr;
                TimeSeriesCollection db(o.rfileCollection, 1);
                int length = db.sequences[0].sequence.size();
                int noSeries = db.sequences.size();
                long int from = 0;
                long int reading_length = length;
                //query the collection
                double milliseconds = 0;
                for(int i = 0; i < noSeries; i++)
                {
                    TimeSeries &query = db.sequences[i];
                    Euclidean euclideanTrue(db, true);
                    Euclidean euclidean(db);
                    euclideanTrue.rangeQuery(query, topn_threshold);
                    double euclideanThreshold = euclidean.distance(query, db.sequences[euclideanTrue.largestDistanceId-1]);

                    QueryEngine qe(db, from, reading_length, prob_threshold, euclideanThreshold, i);
                    qe.ROOT_FOLDER = resultPath;
                    time_duration single = qe.query();
                    milliseconds += single.total_milliseconds();
                }
                milliseconds = milliseconds / noSeries;
                double seconds = (double) milliseconds / (double) 1000;
                //calculate precision recall
                PrecisionRecall pr(noSeries, topn_threshold, resultPath, euclidean_path);
                pr.calculate();
                line += lexical_cast<std::string>(pr.precision) + " ";
                line += lexical_cast<std::string>(pr.recall) + " ";
                line += lexical_cast<std::string>(pr.f1) + " ";
                line += lexical_cast<std::string>(seconds);
                pr_file << line;
                pr_file << endl;
                benchmark.end(true);
            }
            pr_file.close();
        }
    }
    benchmarkall.end(true);
}

void help() {

}

int getdir (string dir, vector<string> &files)
{
    DIR *dp;
    struct dirent *dirp;
    if((dp  = opendir(dir.c_str())) == NULL) {
        cout << "Error(" << errno << ") opening " << dir << endl;
        return errno;
    }
    while ((dirp = readdir(dp)) != NULL) {
    	string name;
    	name.assign(dirp->d_name);
    	if (name != ".." && name != "." && name != ".svn" && name != ".DS_Store")
            files.push_back(string(dirp->d_name));
    }
    closedir(dp);
    return 0;
}

map<string, double> readThresholds()
{
    map<string, double> th_map;
    ifstream thresholdfile (o.rfileThresholds.c_str());
    string line;
    if (thresholdfile.is_open())
    {
        while ( thresholdfile.good() )
        {
            getline (thresholdfile,line);
            if (line.length() == 0)
                continue;
            vector<string> data;
            boost::split(data, line, boost::is_any_of(" "));
            double threshold = lexical_cast<double>(data[1]);
            th_map.insert(std::pair<string, double>(data[0], threshold));
        }
    }
    return th_map;
}

bool FolderExists(string file)
{
    struct stat sb;
    if (stat(file.c_str(), &sb) == 0 && S_ISDIR(sb.st_mode))
    {
        return true;
    }
    return false;
}

void find_max_tao()
{
    vector<string> distributions;
    distributions.push_back("uniform");
    distributions.push_back("normal");
    distributions.push_back("exponential");
    string folder = "/Users/besa/Thesis/udatasets/";
    string fixed_var = "1.0";
    string base_input_path = "/Users/besa/Documents/workspace/proud/data/results/fixed_var_" + fixed_var + "/";
    vector<string> folders = vector<string>();
    getdir(folder.c_str(), folders);
    for(unsigned int d = 0; d < distributions.size(); d++)
    {
        cout << distributions[d] << endl;
        string line = "FOLDER C_LENGTH TS_LENGTH MAX_TAO";
        cout << line << endl;
        for(unsigned int f = 0; f < folders.size(); f++)
        {
            string file_for_folder = folder + folders[f] + "/" + folders[f] + "_dust_exponential_0.4_.dat";
            TimeSeriesCollection db(file_for_folder.c_str(), 1);
            int lengthCollection = db.sequences.size();
            int lengthTs = db.sequences[0].sequence.size();
            string filename = base_input_path + folders[f] + "_proud_" + distributions[d] + "_fixed_var.dat";
            fstream resultsfile (filename.c_str());
            string line_in;
            string max_tao = "";
            double max_f1 = -1;
            if (resultsfile.is_open())
            {
                while ( resultsfile.good() )
                {
                    getline (resultsfile,line_in);
                    if (line_in.length() == 0)
                        continue;
                    vector<string> data;
                    boost::split(data, line_in, boost::is_any_of(" "));
                    string curr_tao = data[2];
                    double curr_f1 = lexical_cast<double> (data[5]);
                    if (curr_f1 > max_f1)
                    {
                        max_f1 = curr_f1;
                        max_tao = curr_tao;
                    }
                }
            }
            resultsfile.close();
            line = folders[f] + " " +
                   lexical_cast<std::string> (lengthCollection) + " " +
                   lexical_cast<std::string> (lengthTs) + " " +
                   max_tao;
            cout << line << endl;
        }
    }
}



/* from here to eof some stuff by michele. */

#define STDDEV_STEP 0.2
#define STDDEV_BEGIN 0.0
#define STDDEV_END 2.0
#define STDDEV_STEPS (int)(((STDDEV_END-STDDEV_BEGIN)/STDDEV_STEP)+1)


const char *names[] = {"50words_t60x6",
                       "Adiac_t60x6",
                       "Beef_t60x6",
                       "CBF_t60x6",
                       "Coffee_t60x6",
                       "ECG200_t60x6",
                       "FaceAll_t60x6",
                       "FaceFour_t60x6",
                       "FISH_t60x6",
                       "Gun_Point_t60x6",
                       "Lighting2_t60x6",
                       "Lighting7_t60x6",
                       "OliveOil_t60x6",
                       "OSULeaf_t60x6",
                       "SwedishLeaf_t60x6",
                       "synthetic_control_t60x6",
                       "Trace_t60x6",
                       NULL};

/*
    const char *names[] = {"50words",
                           "Adiac",
                           "Beef",
                           "CBF",
                           "Coffee",
                           "ECG200",
                           "FaceAll",
                           "FaceFour",
                           "FISH",
                           "Gun_Point",
                           "Lighting2",
                           "Lighting7",
                           "OliveOil",
                           "OSULeaf",
                           "SwedishLeaf",
                           "synthetic_control",
                           "Trace",
                           NULL};


*/
bool active[] = {
    false, //    "50words",
    false, //   "Adiac",
    false, //   "Beef",
    false, //   "CBF",
    false, //   "Coffee",
    false, //   "ECG200",
    false, //   "FaceAll",
    false, //   "FaceFour",
    false, //   "FISH",
    true, //   "Gun_Point",
    false, //    "Lighting2",
    false, //    "Lighting7",
    false, //   "OliveOil",
    false, //  "OSULeaf",
    false, //   "SwedishLeaf",
    false,  //   "synthetic_control",
    false //  "Trace", // 17 datasets
};



void michele() {


    //   for(double probabThreshold = 0.05; probabThreshold <= 0.95+0.1/2.0; probabThreshold+= 0.1) {
    //       experiment4(probabThreshold);
    //   }

    //experiment4(0.45);
    //experiment4(0.05);
    //experiment6(0.45);

    experiment4(0.45);


    exit(0);
}

void experiment4(double probabThreshold) {
    char inputPathname[1024];
    char outputPathname[1024];

    const char *distribs[] = { "uniform", "normal", "exponential", NULL};

    for(int i = 0; names[i]; i++) {
        if(!active[i])
            continue;

        vector<PairIdDistance> topIdDists;
        bool initializeIdDists = true;


        for(int j = 0; distribs[j]; j++) {

            sprintf(outputPathname, "results/%s_proud_%s_%.2f_.dat", names[i], distribs[j], probabThreshold);

            cout << "output file: " << outputPathname << "\n";
            ofstream fout(outputPathname);
            if(!fout) {
                FATAL("error opening outputfile");
            }
            fout.precision(4);

            for(double stddev = STDDEV_BEGIN; stddev <= STDDEV_END+STDDEV_STEP/2.0; stddev +=STDDEV_STEP) {

                cout << "processing dataset:" << names[i] << " distrib:" << distribs[j] << " stddev:" << stddev << endl;
                sprintf(inputPathname, "udatasets_truncated/%s/%s_dust_%s_%.1f_.dat", names[i], names[i], distribs[j], stddev);

                TimeSeriesCollection db(inputPathname, j+1);

                cout << "input file: " << inputPathname << "\n";

                if(initializeIdDists) {
                    db.computeTopKthresholds(10, topIdDists);
                    initializeIdDists = false;
                }

                PROUD proud(db);

                PrecisionRecallM pr("PROUD");
                PrecisionRecallM prEuclidean("Euclidean");

                experiment4_evaluate(db, proud, topIdDists, pr, prEuclidean, probabThreshold);
                fout << j << " " << stddev << " " << pr.getPrecision() <<  " " << pr.getRecall() << " " << pr.getF1() << " " << pr.getTime() << " " <<
                        prEuclidean.getPrecision() << " " << prEuclidean.getRecall() << " " << prEuclidean.getF1() << " "  << prEuclidean.getTime() << endl;

                fout.flush();
            }

            fout.close();

        }
    }
}


void experiment4_evaluate(TimeSeriesCollection &db, PROUD &proud, vector<PairIdDistance> &topIdDists, PrecisionRecallM &pr, PrecisionRecallM &prEuclidean, double probabThreshold) {
    Euclidean euclideanTrue(db, true);
    Euclidean euclidean(db);

    /*
    vector<int> fiveRandomIds;
    fiveRandomIds.push_back(56);
    fiveRandomIds.push_back(12);
    fiveRandomIds.push_back(40);
    fiveRandomIds.push_back(21);
    fiveRandomIds.push_back(30);

    for(unsigned int i = 0; i < 5; i++) {
        TimeSeries &query = db.sequences[fiveRandomIds[i]-1];
*/

    for(unsigned int i = 0; i < db.sequences.size(); i++) {
        TimeSeries &query = db.sequences[i];

        //

        double threshold = topIdDists[query.id-1].second;

        vector<int> exact = euclideanTrue.rangeQuery(query, threshold);
        vector<int> r;

        double euclideanThreshold = euclidean.distance(query, db.sequences[topIdDists[query.id-1].first-1]);

        pr.addStartTime();
        r = proud.rangeQuery(query, euclideanThreshold, probabThreshold);
        pr.addStopTime();
        pr.add(exact, r);
        assert(!isnan(pr.getF1()));

        prEuclidean.addStartTime();
        r = euclidean.rangeQuery(query, euclideanThreshold);
        prEuclidean.addStopTime();
        prEuclidean.add(exact, r);


        assert(!isnan(prEuclidean.getF1()));
    }
}



void experiment6(double probabThreshold) {
    char inputPathname[1024];
    char outputPathname[1024];

    for(int i = 0; names[i]; i++) {
        //     if(!active[i])
        //       continue;

        vector<PairIdDistance> topIdDists;
        bool initializeIdDists = true;


        sprintf(outputPathname, "results/%s_proud_normal_mixed.dat", names[i]);

        cout << "output file: " << outputPathname << "\n";
        ofstream fout(outputPathname);
        if(!fout) {
            FATAL("error opening outputfile");
        }
        fout.precision(4);

        cout << "processing dataset:" << names[i] << " distrib: .. stddev: mixed\n";
        sprintf(inputPathname, "udatasets/%s/%s_dust_normal_mixed_.dat", names[i], names[i]);

        TimeSeriesCollection db(inputPathname, 2); // XXX fixing distribution to normal.

        cout << "input file: " << inputPathname << "\n";

        if(initializeIdDists) {
            db.computeTopKthresholds(10, topIdDists);
            initializeIdDists = false;
        }

        PROUD proud(db);

        PrecisionRecallM pr("PROUD");
        PrecisionRecallM prEuclidean("Euclidean");

        experiment4_evaluate(db, proud, topIdDists, pr, prEuclidean, probabThreshold);
        fout << -1 << " " << "-1" << " " << pr.getPrecision() <<  " " << pr.getRecall() << " " << pr.getF1() << " " << pr.getTime() << " " <<
                prEuclidean.getPrecision() << " " << prEuclidean.getRecall() << " " << prEuclidean.getF1() << " "  << prEuclidean.getTime() << endl;

        fout.flush();
        fout.close();

    }
}
